import os
import json
import codecs
import shlex
import shutil
import glob
import subprocess



CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
MSDIR = os.environ["MSDIR"]
OUTPUT = os.environ["OUTPUT"]

with codecs.open(CONFIG, "r", "utf8") as stdr:
    cab = json.load(stdr)

junk = cab["junk"]
args = []
msname = None

sofia_file = "sofia_parameters.par"
wstd = open(sofia_file, "w")

wstd.write("writeCat.outputDir={:s}\n".format(OUTPUT))
port2tigger = False
image = None
writecat = False
parameterise = False

for param in cab["parameters"]:
    name = param["name"]
    value = param["value"]

    if value is None:
        continue
    if name == "port2tigger":
        port2tigger = value
        continue
    if name == "steps.doWriteCat":
        writecat = value
    if name == "steps.doParameterise":
        parameterise = value
    if name == "import.inFile":
        image = value

    wstd.write("{0}={1}\n".format(name, value))

wstd.close()

_runc = " ".join(["sofia_pipeline.py", sofia_file])

try:
    subprocess.check_call(shlex.split(_runc))
finally:
    for item in junk:
        for dest in [
            OUTPUT,
            MSDIR,
        ]:  # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)
                # Leave other types



import stimela
import os
import unittest
import shutil
import glob
from stimela.exceptions import PipelineException
from stimela.dismissable import dismissable as sdm
from stimela.pathformatter import pathformatter as spf


class TestBasicRecipe(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        unittest.TestCase.setUpClass()
        DIR = os.path.dirname(__file__)
        global INPUT, MSDIR, OUTPUT, MS, PREFIX, LSM, MS_SIM
        INPUT = os.path.join(DIR, "input")
        MSDIR = "msdir"
        global OUTPUT
        OUTPUT = "/tmp/output"
        global CABPATH
        CABPATH = os.path.join(DIR, "cab")

    @classmethod
    def tearDownClass(cls):
        pass

    def tearDown(self):
        unittest.TestCase.tearDown(self)
        global MSDIR
        global OUTPUT
        if os.path.exists(MSDIR):
            shutil.rmtree(MSDIR)
        if os.path.exists(OUTPUT):
            shutil.rmtree(OUTPUT)
        if os.path.exists(INPUT):
            shutil.rmtree(INPUT)
        if os.path.exists("stimela_parameter_files"):
            shutil.rmtree("stimela_parameter_files")
        for log in glob.glob("log-*"):
            os.remove(log)

    def setUp(self):
        unittest.TestCase.setUp(self)
        if not os.path.isdir(INPUT):
            os.mkdir(INPUT)

    def test_define_cab(self):
        global MSDIR
        global INPUT
        global OUTPUT
        global CABPATH
        stimela.register_globals()
        rrr = stimela.Recipe("customcab", ms_dir=MSDIR)
        rrr.add(
            "cab/custom",
            "test1",
            {"bla1": "a"},
            cabpath=CABPATH,
            input=INPUT,
            output=OUTPUT,
        )
        assert len(rrr.jobs) == 1
        rrr.run()  # validate and run
        assert rrr.jobs[0].job._cab.parameters[0].value == "a"
        assert len(rrr.completed) == 1
        assert len(rrr.remaining) == 0

    def test_invalid_choice(self):
        global MSDIR
        global INPUT
        global OUTPUT
        global CABPATH
        stimela.register_globals()
        rrr = stimela.Recipe("invchoice", ms_dir=MSDIR)
        rrr.add(
            "cab/custom",
            "test1",
            {
                "bla1": "d"  # only accepts a, b or c
            },
            cabpath=CABPATH,
            input=INPUT,
            output=OUTPUT,
        )
        with self.assertRaises(PipelineException):
            rrr.run()  # validate and run

    def test_dismissable(self):
        global MSDIR
        global INPUT
        global OUTPUT
        global CABPATH
        stimela.register_globals()
        rrr = stimela.Recipe("testdismissable", ms_dir=MSDIR)
        rrr.add(
            "cab/custom",
            "test1",
            {
                "bla1": "a",  # only accepts a, b or c
                "bla4": sdm("abc"),
                "bla3": sdm(None),
            },
            cabpath=CABPATH,
            input=INPUT,
            output=OUTPUT,
        )
        rrr.run()  # validate and run
        assert rrr.jobs[0].job._cab.parameters[0].value == "a"
        assert rrr.jobs[0].job._cab.parameters[1].value is None
        assert rrr.jobs[0].job._cab.parameters[2].value is None
        assert rrr.jobs[0].job._cab.parameters[3].value == ["abc"]

    def test_floattypefail(self):
        global MSDIR
        global INPUT
        global OUTPUT
        global CABPATH
        stimela.register_globals()
        rrr = stimela.Recipe("testfloattypefail", ms_dir=MSDIR)
        rrr.add(
            "cab/custom",
            "test1",
            {
                "bla1": "a",  # only accepts a, b or c
                "bla3": "1.0a",
            },
            cabpath=CABPATH,
            input=INPUT,
            output=OUTPUT,
        )
        with self.assertRaises(PipelineException):
            rrr.run()  # validate and run

    def test_floattypesuccess(self):
        global MSDIR
        global INPUT
        global OUTPUT
        global CABPATH
        stimela.register_globals()
        rrr = stimela.Recipe("testfloattypesuccess", ms_dir=MSDIR)
        rrr.add(
            "cab/custom",
            "test1",
            {
                "bla1": "a",  # only accepts a, b or c
                "bla3": 4.0,
            },
            cabpath=CABPATH,
            input=INPUT,
            output=OUTPUT,
        )
        rrr.run()  # validate and run
        assert rrr.jobs[0].job._cab.parameters[2].value == [4.0]

    def test_required(self):
        global MSDIR
        global INPUT
        global OUTPUT
        global CABPATH
        stimela.register_globals()
        rrr = stimela.Recipe("testrequired", ms_dir=MSDIR)
        rrr.add(
            "cab/custom",
            "test1",
            {
                "bla3": 4.0,
            },
            cabpath=CABPATH,
            input=INPUT,
            output=OUTPUT,
        )
        with self.assertRaises(PipelineException):
            rrr.run()  # validate and run

    def test_iooverride(self):
        global MSDIR
        global INPUT
        global CABPATH
        with open(os.path.join(INPUT, "testinput.txt"), "w+"):
            pass
        global OUTPUT
        stimela.register_globals()
        rrr = stimela.Recipe("testiooverrides", ms_dir=MSDIR)
        rrr.add(
            "cab/custom",
            "test1",
            {
                "bla1": "a",  # only accepts a, b or c
                "bla2": "testinput.txt:input",
            },
            cabpath=CABPATH,
            input=INPUT,
            output=OUTPUT,
        )
        rrr.run()  # validate and run
        assert rrr.jobs[0].job._cab.parameters[0].value == "a"
        assert rrr.jobs[0].job._cab.parameters[1].value == os.path.join(
            rrr.jobs[0].job.IODEST["input"], "testinput.txt"
        )

    def test_iopathval(self):
        global MSDIR
        global INPUT
        global OUTPUT
        global CABPATH
        stimela.register_globals()
        rrr = stimela.Recipe("ioval", ms_dir=MSDIR)
        rrr.add(
            "cab/custom",
            "test1",
            {
                "bla1": "a",  # only accepts a, b or c
                "bla2": "testinput2.txt:input",
            },
            cabpath=CABPATH,
            input=INPUT,
            output=OUTPUT,
        )
        with self.assertRaises(PipelineException):  # not exist during validation
            rrr.run()  # validate and run

    def test_iopathlist(self):
        global MSDIR
        global INPUT
        global OUTPUT
        global CABPATH
        stimela.register_globals()
        rrr = stimela.Recipe("pathlist", ms_dir=MSDIR)
        rrr.add(
            "cab/custom",
            "test1",
            {
                "bla1": "a",  # only accepts a, b or c
                "bla5": [
                    "testinput2.txt:input",
                    "testinput3.txt:msfile",
                    spf(
                        "{}hello\{reim\}.fits,{}to.fits,{}world.fits",
                        "input",
                        "msfile",
                        "output",
                    ),
                ],
            },
            cabpath=CABPATH,
            input=INPUT,
            output=OUTPUT,
        )
        rrr.run()  # validate and run
        assert rrr.jobs[0].job._cab.parameters[4].value[0] == os.path.join(
            rrr.jobs[0].job.IODEST["input"], "testinput2.txt"
        )
        assert rrr.jobs[0].job._cab.parameters[4].value[1] == os.path.join(
            rrr.jobs[0].job.IODEST["msfile"], "testinput3.txt"
        )
        assert rrr.jobs[0].job._cab.parameters[4].value[
            2
        ] == "{}/hello{{reim}}.fits,{}/to.fits,{}/world.fits".format(
            rrr.jobs[0].job.IODEST["input"],
            rrr.jobs[0].job.IODEST["msfile"],
            rrr.jobs[0].job.IODEST["output"],
        )


if __name__ == "__main__":
    unittest.main()



import glob
import os
import shutil
import unittest

import stimela
from stimela.pathformatter import pathformatter as spf
from stimela import cargo, singularity


class TestBasicRecipe(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        unittest.TestCase.setUpClass()
        global INPUT, MSDIR, OUTPUT, MS, PREFIX, LSM, MS_SIM
        INPUT = os.path.join(os.path.dirname(__file__), "input")
        MSDIR = "msdir"
        global OUTPUT
        OUTPUT = "/tmp/output"

        # Start stimela Recipe instance

        os.chdir(os.path.dirname(__file__))
        cargo.cab.CabDefinition(parameter_file="cab/custom/parameters.json")
        global SINGULARITY, PODMAN
        SINGULARITY = False
        PODMAN = False
        if singularity.version:
            if singularity.BINARY_NAME == "singularity":
                SINGULARITY = singularity.version >= "2.6.0"
            else:
                SINGULARITY = True

    @classmethod
    def tearDownClass(cls):
        pass

    def tearDown(self):
        unittest.TestCase.tearDown(self)
        global MSDIR
        global OUTPUT
        if os.path.exists(MSDIR):
            shutil.rmtree(MSDIR)
        if os.path.exists(OUTPUT):
            shutil.rmtree(OUTPUT)
        if os.path.exists(INPUT):
            shutil.rmtree(INPUT)
        if os.path.exists("stimela_parameter_files"):
            shutil.rmtree("stimela_parameter_files")
        for log in glob.glob("log-*"):
            os.remove(log)

    def setUp(self):
        unittest.TestCase.setUp(self)
        if not os.path.isdir(INPUT):
            os.mkdir(INPUT)

    def test_singularity(self):
        global MSDIR
        global INPUT
        global OUTPUT
        global SINGULARITY

        if SINGULARITY is False:
            return

        stimela.register_globals()
        rrr = stimela.Recipe(
            "singularitypaths",
            ms_dir=MSDIR,
            JOB_TYPE="singularity",
            cabpath="cab/",
            singularity_image_dir=os.environ["STIMELA_PULLFOLDER"],
            log_dir="logs",
        )
        rrr.add(
            "cab/custom",
            "test1",
            {
                "bla1": "a",  # only accepts a, b or c
                "bla5": [
                    "testinput2.txt:input",
                    "testinput3.txt:msfile",
                    spf(
                        "{}hello\{reim\}.fits,{}to.fits,{}world.fits",
                        "input",
                        "msfile",
                        "output",
                    ),
                ],
            },
            input=INPUT,
            output=OUTPUT,
        )
        rrr.run()  # validate and run

        assert rrr.jobs[0].job._cab.parameters[4].value[0] == os.path.join(
            rrr.jobs[0].job.IODEST["input"], "testinput2.txt"
        )
        assert rrr.jobs[0].job._cab.parameters[4].value[1] == os.path.join(
            rrr.jobs[0].job.IODEST["msfile"], "testinput3.txt"
        )
        assert rrr.jobs[0].job._cab.parameters[4].value[
            2
        ] == "{}/hello{{reim}}.fits,{}/to.fits,{}/world.fits".format(
            rrr.jobs[0].job.IODEST["input"],
            rrr.jobs[0].job.IODEST["msfile"],
            rrr.jobs[0].job.IODEST["output"],
        )

    def test_podman(self):
        global MSDIR
        global INPUT
        global OUTPUT
        global PODMAN
        if PODMAN is False:
            return
        stimela.register_globals()
        rrr = stimela.Recipe(
            "podmanpaths",
            ms_dir=MSDIR,
            JOB_TYPE="podman",
            cabpath="cab/",
            log_dir="logs",
        )
        rrr.add(
            "cab/custom",
            "test1",
            {
                "bla1": "a",  # only accepts a, b or c
                "bla5": [
                    "testinput2.txt:input",
                    "testinput3.txt:msfile",
                    spf(
                        "{}hello\{reim\}.fits,{}to.fits,{}world.fits",
                        "input",
                        "msfile",
                        "output",
                    ),
                ],
            },
            input=INPUT,
            output=OUTPUT,
        )
        rrr.run()  # validate and run
        assert rrr.jobs[0].job._cab.parameters[4].value[0] == os.path.join(
            rrr.jobs[0].job.IODEST["input"], "testinput2.txt"
        )
        assert rrr.jobs[0].job._cab.parameters[4].value[1] == os.path.join(
            rrr.jobs[0].job.IODEST["msfile"], "testinput3.txt"
        )
        assert rrr.jobs[0].job._cab.parameters[4].value[
            2
        ] == "{}/hello{{reim}}.fits,{}/to.fits,{}/world.fits".format(
            rrr.jobs[0].job.IODEST["input"],
            rrr.jobs[0].job.IODEST["msfile"],
            rrr.jobs[0].job.IODEST["output"],
        )


if __name__ == "__main__":
    unittest.main()



import subprocess
import os
from stimela import utils
import stimela
import hashlib
from shutil import which


version = None

for item in ["apptainer", "singularity"]:
    BINARY = which(item)
    BINARY_NAME = item
    if BINARY:
        __version_string = subprocess.check_output([BINARY, "--version"]).decode("utf8")
        version = __version_string.strip().split()[-1]
        if BINARY_NAME == "singularity" and version < "3.0.0":
            suffix = ".img"
        else:
            suffix = ".sif"
        break


class SingularityError(Exception):
    pass


def pull(image, name, docker=True, directory=".", force=False):
    """
    pull an image
    """
    if docker:
        fp = "docker://{0:s}".format(image)
    else:
        fp = image
    if not os.path.exists(directory):
        os.mkdir(directory)

    image_path = os.path.abspath(os.path.join(directory, name))
    if os.path.exists(image_path) and not force:
        stimela.logger().info(
            f"Singularity image already exists at '{image_path}'. To replace it, please re-run with the 'force' option"
        )
    else:
        utils.xrun(
            f"cd {directory} && {BINARY}",
            ["pull", "--force" if force else "", "--name", name, fp],
        )
    return 0


class Container(object):
    def __init__(
        self,
        image,
        name,
        volumes=None,
        logger=None,
        time_out=-1,
        runscript="/singularity",
        environs=None,
        workdir=None,
        execdir=".",
    ):
        """
        Python wrapper to singularity tools for managing containers.
        """

        self.image = image
        self.volumes = volumes or []
        self.environs = environs or []
        self.logger = logger
        self.status = None
        self.WORKDIR = workdir
        self.RUNSCRIPT = runscript
        self.PID = os.getpid()
        self.uptime = "00:00:00"
        self.time_out = time_out
        self.execdir = execdir

        self._env = os.environ.copy()

        hashname = hashlib.md5(name.encode("utf-8")).hexdigest()[:3]
        self.name = hashname if version < "3.0.0" else name

    def add_volume(self, host, container, perm="rw", noverify=False):
        if os.path.exists(host) or noverify:
            if self.logger:
                self.logger.debug(
                    "Mounting volume [{0}] in container [{1}] at [{2}]".format(
                        host, self.name, container
                    )
                )
            host = os.path.abspath(host)
        else:
            raise IOError(
                "Path {0} cannot be mounted on container: File doesn't exist".format(
                    host
                )
            )

        self.volumes.append(":".join([host, container, perm]))

        return 0

    def add_environ(self, key, value):
        self.logger.debug(
            "Adding environ varaible [{0}={1}] in container {2}".format(
                key, value, self.name
            )
        )
        self.environs.append("=".join([key, value]))
        key_ = f"{BINARY_NAME.upper()}ENV_{key}"

        self.logger.debug(
            f"Setting singularity environmental variable {key_}={value} on host"
        )
        self._env[key_] = value

        return 0

    def run(self, *args, output_wrangler=None):
        """
        Run a singularity container instance
        """

        if self.volumes:
            volumes = " --bind " + " --bind ".join(self.volumes)
        else:
            volumes = ""

        if not os.path.exists(self.image):
            self.logger.error(
                f"The image, {self.image}, required to run this cab does not exist."
                " Please run 'stimela pull --help' for help on how to download the image"
            )
            raise SystemExit from None

        self.status = "running"
        self._print(
            "Starting container [{0:s}]. Timeout set to {1:d}. The container ID is printed below.".format(
                self.name, self.time_out
            )
        )
        utils.xrun(
            f"{BINARY} run --workdir {self.execdir} --contain",
            list(args) + [volumes, self.image, self.RUNSCRIPT],
            log=self.logger,
            timeout=self.time_out,
            output_wrangler=output_wrangler,
            env=self._env,
            logfile=self.logfile,
        )

        self.status = "exited"

        return 0

    def _print(self, message):
        if self.logger:
            self.logger.info(message)
        else:
            print(message)

        return 0



import stimela
import os
import unittest


class TestKat7Reduce(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        unittest.TestCase.setUpClass()
        # I/O
        global INPUT
        INPUT = "input"
        global MSDIR
        MSDIR = "msdir"

        global MS
        MS = "kat-7-small.ms"
        global PREFIX
        PREFIX = "kat7_small_LBand"

        # Fields
        global GCAL
        GCAL = "PKS2326-477"
        global TARGET
        TARGET = "1"
        global BPCAL
        BPCAL = "PKS1934-638"

        # Reference antenna
        global REFANT
        REFANT = "0"

        # Calibration tables
        global ANTPOS_TABLE
        ANTPOS_TABLE = PREFIX + ".antpos:output"
        global BPCAL_TABLE
        BPCAL_TABLE = PREFIX + ".B0:output"
        global DELAYCAL_TABLE
        DELAYCAL_TABLE = PREFIX + ".K0:output"
        global GAINCAL_TABLE
        GAINCAL_TABLE = PREFIX + ".G0:output"
        global FLUXSCALE_TABLE
        FLUXSCALE_TABLE = PREFIX + ".fluxscale:output"

        global LABEL
        LABEL = "test_reduction"
        global OUTPUT
        OUTPUT = "output_%s" % LABEL
        global MSCONTSUB
        MSCONTSUB = MS + ".contsub"

        global SPW
        SPW = "0:100~355"
        # Calibration tables
        global LSM0
        LSM0 = PREFIX + ".lsm.html"
        global SELFCAL_TABLE1
        SELFCAL_TABLE1 = PREFIX + ".SF1:output"
        global IMAGE1
        IMAGE1 = PREFIX + "image1:output"
        global MASK1
        MASK1 = PREFIX + "mask1.fits"
        global IMAGE2
        IMAGE2 = PREFIX + "image2:output"
        global nchans
        nchans = 256
        global chans
        chans = [100, 355]

        # Clean-Mask-Clean
        global imname0
        imname0 = PREFIX + "image0"
        global maskname0
        maskname0 = PREFIX + "mask0.fits"
        global maskname01
        maskname01 = PREFIX + "mask01.fits"
        global imname1
        imname1 = PREFIX + "image1"
        global corr_ms
        corr_ms = MS + "-corr.ms"
        global lsm0
        lsm0 = PREFIX + "-LSM0"
        stimela.register_globals()

    @classmethod
    def tearDownClass(cls):
        unittest.TestCase.tearDownClass()

    def tearDown(self):
        unittest.TestCase.tearDown(self)

    def setUp(self):
        unittest.TestCase.setUp(self)

    def test_end_to_end_reduction(self):
        global INPUT, OUTPUT, MSDIR, MS, LABEL
        global \
            GAINCAL_TABLE2, \
            FLUXSCALE_TABLE, \
            GAINCAL_TABLE, \
            DELAYCAL_TABLE, \
            BPCAL_TABLE, \
            ANTPOS_TABLE
        global REFANT, BPCAL, TARGET, GCAL, PREFIX
        global MSCONTSUB, SPW, LSM0, SELFCAL_TABLE1, corr_ms, lsm0
        global \
            IMAGE1, \
            IMAGE2, \
            MASK1, \
            nchans, \
            chans, \
            imname0, \
            maskname0, \
            maskname01, \
            imname1

        recipe = stimela.Recipe(
            "Test reduction script", ms_dir=MSDIR, JOB_TYPE="docker", log_dir="logs"
        )

        recipe.add(
            "cab/casa_listobs",
            "listobs",
            {"vis": MS},
            input=INPUT,
            output=OUTPUT,
            label="listobs:: some stats",
            time_out=300,
        )

        recipe.add(
            "cab/owlcat_plotelev",
            "plotobs",
            {
                "msname": MS,
                "output-name": "obsplot.png",
            },
            input=INPUT,
            output=OUTPUT,
            label="plotobs:: Plot elevation/azimuth vs LST/UTC",
        )

        # It is common for the array to require a small amount of time to settle down at the start of a scan. Consequently, it has
        # become standard practice to flag the initial samples from the start
        # of each scan. This is known as 'quack' flagging
        recipe.add(
            "cab/casa_flagdata",
            "quack_flagging",
            {
                "vis": MS,
                "mode": "quack",
                "quackinterval": 10.0,
                "quackmode": "beg",
            },
            input=INPUT,
            output=OUTPUT,
            label="quack_flagging:: Quack flagging",
            time_out=300,
        )

        # Flag the autocorrelations
        recipe.add(
            "cab/politsiyakat_autocorr_amp",
            "flag_autopower",
            {
                "msname": MS,
                "field": "0,1,2",
                "cal_field": "0,2",
                "nrows_chunk": 5000,
                "scan_to_scan_threshold": 1.5,
                "antenna_to_group_threshold": 4,
                "nio_threads": 1,
                "nproc_threads": 32,
            },
            input=INPUT,
            output=OUTPUT,
            label="flag_autopower",
        )

        recipe.add(
            "cab/casa_flagdata",
            "autocorr_flagging",
            {
                "vis": MS,
                "mode": "manual",
                "autocorr": True,
            },
            input=INPUT,
            output=OUTPUT,
            label="autocorr_flagging:: Autocorrelations flagging",
            time_out=300,
        )

        # Flag bad channels
        recipe.add(
            "cab/casa_flagdata",
            "badchan_flagging",
            {
                "vis": MS,
                "mode": "manual",
                "spw": "0:113~113,0:313~313,0:369~369,0:601~607,0:204~204,0:212~212,0:594~600",
            },
            input=INPUT,
            output=OUTPUT,
            label="badchan_flagging:: Bad Channel flagging",
            time_out=300,
        )

        recipe.add(
            "cab/casa_clearcal",
            "clearcal",
            {"vis": MS, "addmodel": True},
            input=INPUT,
            output=OUTPUT,
            label="clearcal:: casa clearcal",
            time_out=300,
        )

        recipe.add(
            "cab/casa_setjy",
            "set_flux_scaling",
            {
                "vis": MS,
                "field": BPCAL,
                "standard": "Perley-Butler 2010",
                "usescratch": True,
                "scalebychan": True,
            },
            input=INPUT,
            output=OUTPUT,
            label="set_flux_scaling:: Set flux density value for the amplitude calibrator",
            time_out=300,
        )

        recipe.add(
            "cab/casa_bandpass",
            "bandpass_cal",
            {
                "vis": MS,
                "caltable": BPCAL_TABLE,
                "field": BPCAL,
                "refant": REFANT,
                "spw": SPW,
                "solint": "inf",
                "bandtype": "B",
                #                        "opacity"   : 0.0,
                #                        "gaincurve" : False,
            },
            input=INPUT,
            output=OUTPUT,
            label="bandpass_cal:: Bandpass calibration",
            time_out=300,
        )

        recipe.add(
            "cab/ragavi",
            "ragavi_gains_plot_bandpass",
            {
                "table": BPCAL_TABLE,
                "gaintype": "B",
                "htmlname": PREFIX + "_B0_amp_chan",
            },
            input=INPUT,
            output=OUTPUT,
            label="ragavi_gains_plot_bandpass:: Plot bandpass table",
            time_out=1200,
        )

        # display the bandpass solutions. Note that in the plotcal inputs below, the amplitudes are being displayed as a function of
        # frequency channel. The parameter subplot=221 is used to display multiple plots per page (2 plots per page in the y
        # direction and 2 in the x direction). The first two commands below show the amplitude solutions (one per each polarization)
        # and the last two show the phase solutions (one per each polarization). Parameter iteration='antenna' is used to step
        # through separate plots for each antenna.
        recipe.add(
            "cab/casa_plotcal",
            "plot_bandpass_amp_R",
            {
                "caltable": BPCAL_TABLE,
                "poln": "R",
                "xaxis": "chan",
                "yaxis": "amp",
                "field": BPCAL,
                "spw": SPW,
                "subplot": 221,
                "figfile": PREFIX + "-B0-R-amp.png",
            },
            input=INPUT,
            output=OUTPUT,
            label="plot_bandpass_amp_R:: Plot bandpass table. AMP, R",
            time_out=1200,
        )

        # Gain calibration - amplitude and phase - first for BPCAL.
        recipe.add(
            "cab/casa_gaincal",
            "gaincal_bp",
            {
                "vis": MS,
                "caltable": GAINCAL_TABLE,
                "field": "{0:s},{1:s}".format(BPCAL, GCAL),
                "solint": "inf",
                "refant": "",
                "gaintype": "G",
                "calmode": "ap",
                "spw": SPW,
                "solnorm": False,
                "gaintable": [BPCAL_TABLE],
                "interp": ["nearest"],
            },
            input=INPUT,
            output=OUTPUT,
            label="gaincal:: Gain calibration",
            time_out=300,
            version=None,
        )

        # Set fluxscale
        recipe.add(
            "cab/casa_fluxscale",
            "fluxscale",
            {
                "vis": MS,
                "caltable": GAINCAL_TABLE,
                "fluxtable": FLUXSCALE_TABLE,
                "reference": [BPCAL],
                "transfer": [GCAL],
                "save_result": "fluxinfo.pickle",
                "incremental": False,
            },
            input=INPUT,
            output=OUTPUT,
            label="fluxscale:: Set fluxscale",
            time_out=300,
        )

        # Apply calibration to BPCAL
        recipe.add(
            "cab/casa_applycal",
            "applycal_bp",
            {
                "vis": MS,
                "field": BPCAL,
                "gaintable": [BPCAL_TABLE, FLUXSCALE_TABLE],
                "gainfield": ["", "", BPCAL],
                "interp": ["", "", "nearest"],
                "calwt": [False],
                "parang": False,
                "applymode": "calflag",
            },
            input=INPUT,
            output=OUTPUT,
            label="applycal_bp:: Apply calibration to Bandpass Calibrator",
            time_out=1800,
        )

        # Flag the phase
        recipe.add(
            "cab/politsiyakat_cal_phase",
            "flag_calphase",
            {
                "msname": MS,
                "field": ",".join(["0", "1", "2"]),
                "cal_field": ",".join(["0", "2"]),
                "nrows_chunk": 5000,
                "data_column": "CORRECTED_DATA",
                "scan_to_scan_threshold": 1.5,
                "baseline_to_group_threshold": 4,
                "nio_threads": 1,
                "nproc_threads": 32,
            },
            input=INPUT,
            output=OUTPUT,
            label="flag_calphase",
        )

        recipe.run()

        recipe = stimela.Recipe(
            "KAT reduction script 2", ms_dir=MSDIR, JOB_TYPE="docker", log_dir="logs"
        )
        # Copy CORRECTED_DATA to DATA, so we can start uv_contsub
        recipe.add(
            "cab/msutils",
            "move_corrdata_to_data",
            {
                "command": "copycol",
                "msname": MS,
                "fromcol": "CORRECTED_DATA",
                "tocol": "DATA",
            },
            input=INPUT,
            output=OUTPUT,
            label="move_corrdata_to_data::msutils",
            time_out=1800,
        )

        os.system("rm -rf {}/{}-corr.ms".format(MSDIR, MS[:-3]))
        recipe.add(
            "cab/casa_split",
            "split_corr_data",
            {
                "vis": MS,
                "outputvis": MS[:-3] + "-corr.ms",
                "field": str(BPCAL),
                "spw": SPW,
                "datacolumn": "data",
            },
            input=INPUT,
            output=OUTPUT,
            label="split_corr_data:: Split corrected data from MS",
            time_out=1800,
        )

        MS = MS[:-3] + "-corr.ms"

        recipe.add(
            "cab/casa_clearcal",
            "prep_split_data",
            {"vis": MS, "addmodel": True},
            input=INPUT,
            output=OUTPUT,
            label="prep_split_data:: Prep split data with casa clearcal",
            time_out=1800,
        )

        # Clean-Mask-Clean
        imname0 = PREFIX + "image0"
        maskname0 = PREFIX + "mask0.fits"
        maskname01 = PREFIX + "mask01.fits"
        imname1 = PREFIX + "image1"

        recipe.add(
            "cab/casa_tclean",
            "image_target_field_r1",
            {
                "vis": MS,
                "datacolumn": "corrected",
                "field": "0",
                "start": 21,  # Other channels don't have any data
                "nchan": 235 - 21,
                "width": 1,
                # Use Briggs weighting to weigh visibilities for imaging
                "weighting": "briggs",
                "robust": 0,
                "imsize": 256,  # Image size in pixels
                "cellsize": "30arcsec",  # Size of each square pixel
                "niter": 100,
                "stokes": "I",
                "prefix": "%s:output" % (imname1),
            },
            input=INPUT,
            output=OUTPUT,
            label="image_target_field_r1:: Image target field second round",
            time_out=300,
        )

        recipe.add(
            "cab/cleanmask",
            "mask0",
            {
                "image": "%s.image.fits:output" % (imname1),
                "output": "%s:output" % (maskname0),
                "dilate": False,
                "sigma": 20,
            },
            input=INPUT,
            output=OUTPUT,
            label="mask0:: Make mask",
            time_out=1800,
        )

        lsm0 = PREFIX + "-LSM0"
        # Source finding for initial model
        recipe.add(
            "cab/pybdsm",
            "extract_init_model",
            {
                "image": "%s.image.fits:output" % (imname1),
                "outfile": "%s:output" % (lsm0),
                "thresh_pix": 25,
                "thresh_isl": 15,
                "port2tigger": True,
            },
            input=INPUT,
            output=OUTPUT,
            label="extract_init_model:: Make initial model from preselfcal image",
            time_out=1800,
        )

        # First selfcal round

        recipe.add(
            "cab/calibrator",
            "calibrator_Gjones_subtract_lsm0",
            {
                "skymodel": "%s.lsm.html:output" % (lsm0),
                "msname": MS,
                "threads": 16,
                "column": "DATA",
                "output-data": "CORR_RES",
                "Gjones": True,
                # Ad-hoc right now, subject to change
                "Gjones-solution-intervals": [20, 0],
                "Gjones-matrix-type": "GainDiagPhase",
                "tile-size": 512,
                "field-id": 0,
            },
            input=INPUT,
            output=OUTPUT,
            label="calibrator_Gjones_subtract_lsm0:: Calibrate and subtract LSM0",
            time_out=1800,
        )

        # Diversity is a good thing... lets add some DDFacet to this soup bowl
        imname = PREFIX + "ddfacet"

        recipe.add(
            "cab/ddfacet",
            "ddfacet_test",
            {
                "Data-MS": [MS],
                "Output-Name": imname,
                "Image-NPix": 256,
                "Image-Cell": 30,
                "Cache-Reset": True,
                "Freq-NBand": 2,
                "Weight-ColName": "WEIGHT",
                "Data-ChunkHours": 10,
                "Beam-FITSFeed": "rl",
                "Data-Sort": True,
                "Log-Boring": True,
                "Deconv-MaxMajorIter": 1,
                "Deconv-MaxMinorIter": 20,
            },
            input=INPUT,
            output=OUTPUT,
            shared_memory="8gb",
            label="image_target_field_r0ddfacet:: Make a test image using ddfacet",
            time_out=520,
        )

        lsm1 = PREFIX + "-LSM0"
        # Source finding for initial model
        recipe.add(
            "cab/pybdsm",
            "extract_init_model",
            {
                "image": "%s.app.restored.fits:output" % (imname),
                "outfile": "%s:output" % (lsm1),
                "thresh_pix": 25,
                "thresh_isl": 15,
                "port2tigger": True,
            },
            input=INPUT,
            output=OUTPUT,
            label="extract_init_model:: Make initial model from preselfcal image",
            time_out=1800,
        )

        # Stitch LSMs together
        lsm2 = PREFIX + "-LSM2"
        recipe.add(
            "cab/tigger_convert",
            "stitch_lsms1",
            {
                "input-skymodel": "%s.lsm.html:output" % lsm0,
                "output-skymodel": "%s.lsm.html:output" % lsm2,
                "rename": True,
                "force": True,
                "append": "%s.lsm.html:output" % lsm1,
            },
            input=INPUT,
            output=OUTPUT,
            label="stitch_lsms1::Create master lsm file",
            time_out=300,
        )

        recipe.add(
            "cab/calibrator",
            "calibrator_Gjones_subtract_lsm0",
            {
                "skymodel": "%s.lsm.html:output" % (lsm2),
                "msname": MS,
                "threads": 16,
                "column": "DATA",
                "output-data": "CORR_RES",
                "Gjones": True,
                # Ad-hoc right now, subject to change
                "Gjones-solution-intervals": [20, 0],
                "Gjones-matrix-type": "GainDiagPhase",
                "tile-size": 512,
                "field-id": 0,
            },
            input=INPUT,
            output=OUTPUT,
            label="calibrator_Gjones_subtract_lsm0:: Calibrate and subtract LSM0",
            time_out=1800,
        )

        recipe.add(
            "cab/casa_uvcontsub",
            "uvcontsub",
            {
                "msname": MS,
                "field": "0",
                "fitorder": 1,
            },
            input=INPUT,
            output=OUTPUT,
            label="uvcontsub:: Subtract continuum in the UV plane",
            time_out=1800,
        )

        # Image HI
        recipe.add(
            "cab/casa_clean",
            "casa_dirty_cube",
            {
                "msname": MS + ".contsub",
                "prefix": PREFIX,
                "mode": "channel",
                "nchan": nchans,
                "niter": 0,
                "npix": 256,
                "cellsize": 30,
                "weight": "natural",
                "port2fits": True,
            },
            input=INPUT,
            output=OUTPUT,
            label="casa_dirty_cube:: Make a dirty cube with CASA CLEAN",
            time_out=1800,
        )

        recipe.add(
            "cab/sofia",
            "sofia",
            {
                #    USE THIS FOR THE WSCLEAN DIRTY CUBE
                #    "import.inFile"     :   '{:s}-cube.dirty.fits:output'.format(combprefix),
                #    USE THIS FOR THE CASA CLEAN CUBE
                # CASA CLEAN cube
                "import.inFile": "{:s}.image.fits:output".format(PREFIX),
                "steps.doFlag": False,
                "steps.doScaleNoise": True,
                "steps.doSCfind": True,
                "steps.doMerge": True,
                "steps.doReliability": False,
                "steps.doParameterise": False,
                "steps.doWriteMask": True,
                "steps.doMom0": False,
                "steps.doMom1": False,
                "steps.doWriteCat": True,
                "flag.regions": [],
                "scaleNoise.statistic": "mad",
                "SCfind.threshold": 4,
                "SCfind.rmsMode": "mad",
                "merge.radiusX": 2,
                "merge.radiusY": 2,
                "merge.radiusZ": 2,
                "merge.minSizeX": 2,
                "merge.minSizeY": 2,
                "port2tigger": False,
                "merge.minSizeZ": 2,
            },
            input=INPUT,
            output=OUTPUT,
            label="sofia:: Make SoFiA mask and images",
            time_out=1800,
        )

        recipe.run()



import os
import yaml
import sharpener
import glob
import shlex
import subprocess
import shutil


CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
MSDIR = os.environ["MSDIR"]
OUTPUT = os.environ["OUTPUT"]

with open(CONFIG, "r") as _std:
    cab = yaml.safe_load(_std)

junk = cab["junk"]
args = []
msname = None

pkg_path = os.path.dirname(os.path.realpath(sharpener.__file__))
sharpener_file = "{:s}/sharpener_default.yml".format(pkg_path)

with open(sharpener_file) as f:
    list_doc = yaml.safe_load(f)

for param in cab["parameters"]:
    name = param["name"]
    value = param["value"]

    if value is None:
        continue

    for key, val in list_doc.items():
        if isinstance(val, dict):
            for k1, v1 in val.items():
                if "enable" in name:
                    if key in name:
                        list_doc[key]["enable"] = value
                elif k1 == name:
                    list_doc[key][k1] = value
        else:
            if key == name:
                list_doc[key] = value

# Get the relative path from workdir
list_doc["general"]["contname"] = os.path.relpath(
    list_doc["general"]["contname"], list_doc["general"]["workdir"]
)
list_doc["general"]["cubename"] = os.path.relpath(
    list_doc["general"]["cubename"], list_doc["general"]["workdir"]
)
list_doc["source_catalog"]["catalog_file"] = os.path.relpath(
    list_doc["source_catalog"]["catalog_file"], list_doc["general"]["workdir"]
)

edited_file = "sharpener_default.yml"
with open(edited_file, "w") as f:
    yaml.dump(list_doc, f)

_runc = "run_sharpener -c %s" % edited_file
try:
    subprocess.check_call(shlex.split(_runc))
finally:
    for item in junk:
        for dest in [
            OUTPUT,
            MSDIR,
        ]:  # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)
                # Leave other types



import sys
import os
from MSUtils import msutils
import MSUtils.ClassESW as esw
import inspect
from MSUtils.imp_plotter import gain_plotter
import glob
import shutil
import yaml


CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

with open(CONFIG, "r") as _std:
    cab = yaml.safe_load(_std)

junk = cab["junk"]
args = {}
for param in cab["parameters"]:
    name = param["name"]
    value = param["value"]

    if name == "command":
        function = value
        continue

    args[name] = value

if function == "sumcols":
    args["outcol"] = args.pop("colname")

if function == "estimate_weights":
    msnoise = esw.MSNoise(args["msname"])
    if (
        isinstance(args["stats_data"], str)
        and args["stats_data"].find("use_package_meerkat_spec") >= 0
    ):
        args["stats_data"] = esw.MEERKAT_SEFD

    # Calculate noise/weights from spec
    noise, weights = msnoise.estimate_weights(
        stats_data=args["stats_data"],
        smooth=args["smooth"],
        fit_order=args["fit_order"],
        plot_stats=args.get("plot_stats", None),
    )

    if args["write_to_ms"]:
        msnoise.write_toms(noise, columns=args["noise_columns"])
        msnoise.write_toms(weights, columns=args["weight_columns"])
    sys.exit(0)

if function == "plot_gains":
    tab = args["ctable"]
    tabtype = args["tabtype"]
    dpi = args["plot_dpi"]
    scale = args["subplot_scale"]
    outfile = args["plot_file"]
    gain_plotter(tab, tabtype, outfile, scale, dpi)
    sys.exit(0)


run_func = getattr(msutils, function, None)
if run_func is None:
    raise RuntimeError("Function '{}' is not part of MSUtils".format(function))

# Filter default parameters that are part of this function
func_args = inspect.getargspec(run_func)[0]
_args = {}
for arg in args.keys():
    if arg in func_args:
        _args[arg] = args[arg]

try:
    run_func(**_args)
finally:
    for item in junk:
        for dest in [
            OUTPUT,
            MSDIR,
        ]:  # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)



# -*- coding: future_fstrings -*-
import Crasa.Crasa as crasa
from scabha import config, parameters_dict

print(f"Running CASA task '{config.binary}'")

makemask_args = {}
immath_args = {}
for name, value in parameters_dict.items():
    if value is None:
        continue

    if name in ["threshold", "inpimage", "output"]:
        if name in ["threshold"]:
            im_value = " iif( IM0 >=%s, IM0, 0.0) " % value
            im_name = "expr"
        if name in ["output"]:
            im_value = "%s_thresh" % value
            im_name = "outfile"
        if name in ["inpimage"]:
            im_value = value
            im_name = "imagename"
        immath_args[im_name] = im_value

    if name in ["mode", "inpimage", "inpmask", "output", "overwrite"]:
        makemask_args[name] = value

if "expr" in immath_args:
    task = crasa.CasaTask("immath", **immath_args)
    task.run()

if "inpmask" not in makemask_args:
    makemask_args["inpmask"] = immath_args["outfile"]

task = crasa.CasaTask(config.binary, **makemask_args)
task.run()



import os
import subprocess
import shlex
import glob
import shutil
import yaml


CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
MSDIR = os.environ["MSDIR"]
OUTPUT = os.environ["OUTPUT"]

with open(CONFIG, "r") as _std:
    cab = yaml.safe_load(_std)
junk = cab["junk"]

args = []
targets = None
for param in cab["parameters"]:
    name = param["name"]
    value = param["value"]

    if value is None:
        continue
    elif value is False:
        continue
    elif value is True:
        value = ""
    elif name == "target-images":
        targets = value
        continue

    args += ["{0}{1} {2}".format(cab["prefix"], name, value)]

indir = os.path.dirname(targets[0])
target_names = map(os.path.basename, targets)
target_images = "--target-images " + " --target-images ".join(target_names)

args += ["--input {0:s} {1:s} --output {2:s}".format(indir, target_images, OUTPUT)]

if not target_images:
    raise RuntimeError(
        "Filenames of the images to be mosaicked have not been specified."
    )
_runc = " ".join([cab["binary"]] + args)

try:
    subprocess.check_call(shlex.split(_runc))
finally:
    for item in junk:
        for dest in [
            OUTPUT,
            MSDIR,
        ]:  # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)



import os
import shutil
import shlex
import subprocess
import glob
import yaml

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

with open(CONFIG, "r") as _std:
    cab = yaml.safe_load(_std)

junk = cab["junk"]

args = []
inimage = None
outimage = None
stack = False
unstack = False
axis = None
chunk = 1
file_pattern = False

for param in cab["parameters"]:
    value = param["value"]
    name = param["name"]

    if value in [None, False]:
        continue

    if name == "image":
        inimage = " ".join(value)
        continue
    elif name == "output":
        outimage = value
        continue
    elif name == "stack":
        stack = True
        continue
    elif name == "unstack":
        unstack = True
        continue
    elif name == "unstack-chunk":
        chunk = value
        continue
    elif name == "fits-axis":
        axis = value
        continue
    elif name == "file_pattern":
        value = '"%s"' % value
        file_pattern = True

    elif value is True:
        value = ""

    args.append("{0}{1} {2}".format(cab["prefix"], name, value))

if stack and axis:
    args.append("{0}stack {1}:{2}".format(cab["prefix"], outimage, axis))
    outimage = None
elif unstack and axis:
    args.append("{0}unstack {1}:{2}:{3}".format(cab["prefix"], outimage, axis, chunk))
    outimage = None
else:
    outimage = "{0}output {1}".format(cab["prefix"], outimage)

if file_pattern:
    inimage = ""

_runc = " ".join([cab["binary"]] + args + [inimage, outimage or ""])

try:
    subprocess.check_call(shlex.split(_runc))
finally:
    for item in junk:
        for dest in [
            OUTPUT,
            MSDIR,
        ]:  # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)



import os
import glob
import subprocess
import yaml
import shutil
import shlex

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

with open(CONFIG, "r") as _std:
    cab = yaml.safe_load(_std)

junk = cab["junk"]

params = {}
for param in cab["parameters"]:
    name = param["name"]
    value = param["value"]

    if value in [False, None]:
        continue

    if value is True:
        value = ""

    # Positional arguments
    if name == "input-skymodel":
        inlsm = value
        continue
    elif name == "tag":
        tag = value
        continue

    params[name] = value

# TODO: Need fix tigger-tag, these kludges are annoying
if params.pop("transfer-tags", False) in [True, ""]:
    if params.get("tolerance", None) is None:
        raise RuntimeError(
            "Parameter 'tolerance' is required when 'transfer-tags' is enables"
        )
    args = [
        "{0}transfer-tags {1}:{2}".format(cab["prefix"], inlsm, params.pop("tolerance"))
    ]
    inlsm = params.get("output")
else:
    args = []

args += [
    "{0}{1} {2}".format(cab["prefix"], name, value)
    for name, value in params.iteritems()
]

_runc = " ".join([cab.binary, inlsm, tag] + args)

try:
    subprocess.check_call(shlex.split(_runc))
finally:
    for item in junk:
        for dest in [
            OUTPUT,
            MSDIR,
        ]:  # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)



import os
import Crasa.Crasa as crasa
import yaml
import glob
import shutil

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

with open(CONFIG, "r") as _std:
    cab = yaml.safe_load(_std)
junk = cab["junk"]

args = {}
for param in cab["parameters"]:
    name = param["name"]
    value = param["value"]

    if value is None:
        continue

    args[name] = value

task = crasa.CasaTask(cab["binary"], **args)
try:
    task.run()
finally:
    for item in junk:
        for dest in [
            OUTPUT,
            MSDIR,
        ]:  # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)
                # Leave other types



import os
import astropy.io.fits as pyfits
import glob
import subprocess
import shutil
import shlex
import yaml


CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

with open(CONFIG, "r") as _std:
    cab = yaml.safe_load(_std)

junk = cab["junk"]

args = []
parset = None
for param in cab["parameters"]:
    name = param["name"]
    value = param["value"]
    if name == "Parset" and value is not None:
        parset = value
        continue
    if name == "Parset" and value is None:
        continue
    if name == "Noise-Image" and value is None:
        continue

    if isinstance(value, list):
        arg = "{0}{1} {2}".format(cab["prefix"], name, ",".join(value))
    else:
        arg = "{0}{1} {2}".format(cab["prefix"], name, value)

    args.append(arg)


removed = False
for item1 in args:
    if "Noise-Image" in item1:
        noise_image = item1.split("{0}Noise-Image ".format(cab["prefix"]))[-1]
        args.remove("{0}Noise-Image {1}".format(cab["prefix"], noise_image))
        noise_hdu = pyfits.open(noise_image)
        noise_data = noise_hdu[0].data
        noise_std = noise_data.std()
        noise_hdu.close()
        for item2 in args:
            if "Noise-Sigma" in item2:
                noise_sigma = item2.split("{0}Noise-Sigma ".format(cab["prefix"]))[-1]
                args.remove("{0}Noise-Sigma {1}".format(cab["prefix"], noise_sigma))
                removed = True
                threshold = float(noise_sigma) * noise_std
                for item3 in args:
                    if "{0}Deconv-FluxThreshold".format(cab["prefix"]) in item3:
                        args.remove(item3)
                args.append(
                    "{0}Deconv-FluxThreshold {1}".format(cab["prefix"], threshold)
                )
if not removed:
    args.remove("{0}Noise-Sigma 3.0".format(cab["prefix"]))

if parset is not None:
    args.insert(0, parset)


_runc = " ".join([cab["binary"]] + args)

try:
    subprocess.check_call(shlex.split(_runc))
finally:
    for item in junk:
        for dest in [
            OUTPUT,
            MSDIR,
        ]:  # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)
                # Leave other types



import os
import sys
import Tigger
import numpy
import tempfile
import json
import codecs
import shlex
import shutil
import glob
import subprocess

from astLib.astWCS import WCS
from astropy.io.votable import parse_single_table
from Tigger.Models import SkyModel, ModelClasses


CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
MSDIR = os.environ["MSDIR"]
OUTPUT = os.environ["OUTPUT"]

with codecs.open(CONFIG, "r", "utf8") as stdr:
    cab = json.load(stdr)

junk = cab["junk"]
args = []
msname = None

sofia_file = "sofia_parameters.par"
wstd = open(sofia_file, "w")

wstd.write("output.directory={:s}\n".format(OUTPUT))
port2tigger = False
image = None

for param in cab["parameters"]:
    name = param["name"]
    value = param["value"]
    dtype = param["dtype"]

    # Fix the sofia issue of needing lowercase booleans.
    if dtype == "bool":
        if value and (not name == "port2tigger"):
            value = "true"
        elif not name == "port2tigger":
            value = "false"

    if dtype == "list:str" and value is not None:
        try:
            value = ",".join(map(str, value))  # Flatten the list into a string
        except ValueError:
            pass

    if value is None:
        continue
    if name == "port2tigger":
        port2tigger = value
        continue
    if name == "output.writeCatXML":
        writecat = value
    if name == "parameter.enable":
        parameterise = value
    if name == "input.data":
        image = value

    wstd.write("{0}={1}\n".format(name, value))

wstd.close()

_runc = " ".join([cab["binary"], sofia_file])

try:
    subprocess.check_call(shlex.split(_runc))
finally:
    for item in junk:
        for dest in [
            OUTPUT,
            MSDIR,
        ]:  # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)
                # Leave other types

if not port2tigger:
    sys.exit(0)

# convert to data file to Tigger LSM
# First make dummy tigger model
tfile = tempfile.NamedTemporaryFile(suffix=".txt")
tfile.flush()

if image and writecat and parameterise:
    pass
else:
    sys.exit(0)

prefix = os.path.splitext(image)[0]
tname_lsm = prefix + ".lsm.html"
with open(tfile.name, "w") as stdw:
    stdw.write("#format:name ra_d dec_d i emaj_s emin_s pa_d\n")

model = Tigger.load(tfile.name)
tfile.close()


def tigger_src(src, idx):
    name = "SRC%d" % idx
    flux = ModelClasses.Polarization(src["f_sum"], 0, 0, 0)
    ra = numpy.deg2rad(src["ra"])
    dec = numpy.deg2rad(src["dec"])
    pos = ModelClasses.Position(ra, dec)
    ex = numpy.deg2rad(src["ell_maj"])
    ey = numpy.deg2rad(src["ell_min"])
    pa = numpy.deg2rad(src["ell_pa"])
    print(name)

    if ex and ey:
        shape = ModelClasses.Gaussian(ex, ey, pa)
    else:
        shape = None
    source = SkyModel.Source(name, pos, flux, shape=shape)
    # Adding source peak flux (error) as extra flux attributes for sources,
    # and to avoid null values for point sources I_peak = src["Total_flux"]
    if shape:
        source.setAttribute("I_peak", float(src["f_max"]))
    else:
        source.setAttribute("I_peak", float(src["f_sum"]))

    return source


table = parse_single_table("{0}_cat.xml".format(prefix))
data = table.array

for i, src in enumerate(data):
    model.sources.append(tigger_src(src, i))

wcs = WCS(image)
centre = wcs.getCentreWCSCoords()
model.ra0, model.dec0 = map(numpy.deg2rad, centre)

model.save(tname_lsm)
# Rename using CORPAT
_runc = "tigger-convert %s --rename -f" % tname_lsm
try:
    subprocess.check_call(shlex.split(_runc))
finally:
    for item in junk:
        for dest in [
            OUTPUT,
            MSDIR,
        ]:  # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)



import pyrap.images
import os
from pyrap.tables import table
from MSUtils import msutils
import tempfile
import astropy.io.fits as pyfits
import subprocess
import shlex
import yaml


CONFIG = os.environ["CONFIG"]
OUTPUT = os.environ["OUTPUT"]
INPPUT = os.environ["INPUT"]
MSDIR = os.environ["MSDIR"]

with open(CONFIG, "r") as _std:
    cab = yaml.safe_load(_std)

junk = cab["junk"]

params = cab["parameters"]

tdir = tempfile.mkdtemp(dir=".")
os.chdir(tdir)


def rm_fr(item):
    os.system("rm -fr {}".format(item))


def _run(prefix=None, predict=False, **kw):
    port2fits = kw.pop("port2fits", True)
    keep_casa_images = kw.pop("keep_casa_images", False)

    if predict:
        args = ["{0}={1}".format(a, b) for a, b in kw.iteritems()]
        _runc = " ".join([cab["binary"]] + args)
        subprocess.check_call(shlex.split(_runc))
        return

    if kw.get("niter", 0) > 0:
        if kw.get("operation", None) not in [
            "clark",
            "hogbom",
            "csclean",
            "multiscale",
            "entropy",
        ]:
            kw["operation"] = "csclean"
        images = {
            "restored": ["{0}.restored.{1}".format(prefix, a) for a in ["fits", "img"]],
            "model": ["{0}.model.{1}".format(prefix, a) for a in ["fits", "img"]],
            "residual": ["{0}.residual.{1}".format(prefix, a) for a in ["fits", "img"]],
        }

    elif kw.get("niter", 0) == 0:
        kw["operation"] = "image"

        images = {
            "image": ["{0}.dirty.{1}".format(prefix, a) for a in ["fits", "img"]],
        }

    for key, value in images.iteritems():
        kw[key] = value[1]

    args = ["{0}={1}".format(a, b) for a, b in kw.iteritems()]
    _runc = " ".join([cab["binary"]] + args)
    subprocess.check_call(shlex.split(_runc))

    if port2fits:
        print("Converting CASA iamges to FITS images")
        for fits, img in images.itervalues():
            im = pyrap.images.image(img)
            im.tofits(fits, overwrite=True, velocity=kw.get("prefervelocity", False))
            if not keep_casa_images:
                rm_fr(img)


def predict_vis(
    msname, image, column="MODEL_DATA", chanchunk=None, chanstart=0, chanstep=1
):
    """Converts image into predicted visibilities"""

    # CASA to convert them
    casaimage = "{0}/{1}.img".format(OUTPUT, os.path.basename(image))

    # convert to CASA image
    img = pyrap.images.image(image)
    img.saveas(casaimage)

    imgshp = img.shape()
    ftab = table(msname + "/SPECTRAL_WINDOW")
    numchans = ftab.getcol("NUM_CHAN")[0]
    # default chunk list is entire chanel range. Update this if needed
    chunklist = [(0, numchans, None, None)]
    if len(imgshp) == 4 and imgshp[0] > 1:
        nimgchan = imgshp[0]

        print(
            "image cube has {0} channels, MS has {1} channels".format(
                nimgchan, numchans
            )
        )

        imgchansize = (
            imgshp[1] * imgshp[2] * imgshp[3] * 4
        )  # size of an image channel in bytes
        if chanchunk is None:
            mem_bytes = os.sysconf("SC_PAGE_SIZE") * os.sysconf(
                "SC_PHYS_PAGES"
            )  # e.g. 4015976448
            chanchunk = max((mem_bytes / 20) / imgchansize, 1)
            print(
                "based on available memory ({0}), max image chunk is {1} channels".format(
                    mem_bytes, chanchunk
                )
            )
        if chanchunk < nimgchan:
            mschanstep = numchans * chanstep / nimgchan
            if numchans % nimgchan:
                print(
                    "MS channels not evenly divisible into $nimgchan image channels, chunking may be incorrect"
                )
            chunklist = []
            for chan0 in range(0, nimgchan, chanchunk):
                imch0, imch1 = chan0, (min(chan0 + chanchunk, nimgchan) - 1)
                msch0 = chanstart + imch0 * mschanstep
                msnch = (imch1 - imch0 + 1) * mschanstep / chanstep
                # overlap each chunk from 1 onwards by a half-chunk back to take care of extrapolated visibilties
                # from previous channel
                if imch0:
                    imch0 -= 1
                    msch0 -= mschanstep / 2
                    msnch += mschanstep / 2
                print(
                    "image chunk {0}~{1} corresponds to MS chunk {2}~{3}".format(
                        imch0, imch1, msch0, msch0 + msnch - 1
                    )
                )
                chunklist.append((msch0, msnch, imch0, imch1))

    # even in fill-model mode where it claims to ignore image parameters, the image channelization
    # arguments need to be "just so" as per below, otherwise it gives a GridFT: weights all zero message
    kw0 = {}
    kw0.update(
        ms=msname,
        model=casaimage,
        niter=0,
        fixed=1,
        mode="channel",
        operation="csclean",
        img_nchan=1,
        img_chanstart=chanstart,
        img_chanstep=numchans * chanstep,
    )
    kw0["fillmodel"] = 1

    blc = [0] * len(imgshp)
    trc = [x - 1 for x in imgshp]

    # now loop over image frequency chunks
    for ichunk, (mschanstart, msnumchans, imgch0, imgch1) in enumerate(chunklist):
        if len(chunklist) > 1:
            blc[0], trc[0] = imgch0, imgch1
            print("writing CASA image for slice {0} {1}".format(blc, trc))
            casaimage1 = "{0}.{1}.img".format(image, ichunk)
            rm_fr(casaimage1)
            print(
                "writing CASA image for slice {0} {1} to {2}".format(
                    blc, trc, casaimage1
                )
            )
            img.subimage(blc, trc, dropdegenerate=False).saveas(casaimage1)
            kw0.update(model=casaimage1)
        else:
            img.unlock()
        # setup imager options
        kw0.update(chanstart=mschanstart, chanstep=chanstep, nchan=msnumchans)
        print("predicting visibilities into MODEL_DATA")

        _run(predict=True, **kw0)
        if len(chunklist) > 1:
            rm_fr(casaimage1)
    rm_fr(casaimage)

    if column != "MODEL_DATA":
        print(
            "Data was predicted to MODEL_DATA column. Will now copy it to the {} column as requested".format(
                column
            )
        )
        msutils.copycol(msname=msname, fromcol="MODEL_DATA", tocol=column)


options = {}
for param in params:
    value = param["value"]
    name = param["name"]

    if name == "prefix":
        prefix = value
        continue

    if value is None:
        continue
    if name == "cellsize":
        if isinstance(value, (float, int)):
            value = "{}arcsec".format(value)
    elif name in ["threshold", "targetflux"]:
        if isinstance(value, float):
            value = "{}arcsec".format(value)
    options[name] = value

noise_image = options.pop("noise_image", False)
if noise_image:
    noise_sigma = options.pop("noise_sigma")
    noise_hdu = pyfits.open(noise_image)
    noise_data = noise_hdu[0].data
    noise_std = noise_data.std()
    threshold = noise_sigma * noise_std
    options["threshold"] = "{}Jy".format(threshold)
else:
    options.pop("noise_sigma")

predict = options.pop("simulate_fits", False)
if predict:
    tfile = tempfile.NamedTemporaryFile(suffix=".fits")
    tfile.flush()
    cell = options.get("cellsize", None)
    if cell is None:
        with pyfits.open(predict) as _hdu:
            if hasattr(_hdu, "__iter__"):
                hdu = _hdu[0]
            else:
                hdu = _hdu

            cdelt = hdu.header.get("CDELT1", None)
            if cdelt:
                cell = "{:f}arcsec".format(abs(cdelt) * 3600)

    if cell is None:
        raise RuntimeError(
            'The size of a pixel in this FITS image was not specified \
in FITS header (CDELT1/2), or as parameter for this module ("cellsize"). Cannot proceed'
        )

    _runc = " ".join(
        ["python /scratch/code/predict_from_fits.py"]
        + [predict, options["ms"], cell, tfile.name]
    )
    subprocess.check_call(shlex.split(_runc))
    predict_vis(
        msname=options["ms"],
        image=tfile.name,
        column=options.get("data", "MODEL_DATA"),
        chanchunk=options.get("chanchunk", None),
        chanstart=options.get("img_chanstart", 0),
        chanstep=options.get("img_chanstep", 1),
    )
    tfile.close()
else:
    _run(prefix, **options)

os.chdir(OUTPUT)
os.system("rm -r {}".format(tdir))



import sys
import os
import codecs
import time
import subprocess

from threading import Event, Thread

from . import StimelaCabRuntimeError

DEBUG = False
INTERRUPT_TIME = 2.0  # seconds -- do not want to constantly interrupt the child process
LIVELOG_TIME = 0.1


def xrun(
    command,
    options,
    log=None,
    _log_container_as_started=False,
    logfile=None,
    timeout=-1,
    kill_callback=None,
):
    """
    Run something on command line.

    Example: _run("ls", ["-lrt", "../"])
    """
    if "LOGFILE" in os.environ and logfile is None:
        logfile = os.environ["LOGFILE"]  # superceed if not set

    # skip lines from previous log files
    if logfile is not None and os.path.exists(logfile):
        with codecs.open(
            logfile, "r", encoding="UTF-8", errors="ignore", buffering=0
        ) as foutlog:
            prior_log_bytes_read = foutlog.tell()
    else:  # not existant, create
        prior_log_bytes_read = 0
        if logfile is not None and not os.path.exists(logfile):
            with codecs.open(
                logfile, "w+", encoding="UTF-8", errors="ignore", buffering=0
            ) as foutlog:
                pass

    cmd = " ".join([command] + list(map(str, options)))

    def _remove_ctrls(msg):
        import re

        ansi_escape = re.compile(r"\x1B[@-_][0-?]*[ -/]*[@-~]")
        return ansi_escape.sub("", msg)

    def _print_info(msg, loglevel="INFO"):
        if msg is None:
            return
        msg = _remove_ctrls(msg)
        if msg.strip() == "":
            return
        if log:
            try:
                getattr(log, loglevel.lower())(msg.rstrip("\n"))
            except UnicodeError:
                log.warn("Log contains unicode and will not be printed")
        else:
            try:
                (print(msg),)
            except UnicodeError:
                print("Log contains unicode and will not be printed")

    def _print_warn(msg):
        if msg is None:
            return
        msg = _remove_ctrls(msg)
        if msg.strip() == "":
            return
        if log:
            try:
                log.info(msg.rstrip("\n"))
            except UnicodeError:
                log.warn("Log contains unicode and will not be printed")
        else:
            try:
                (print(msg),)
            except UnicodeError:
                print("Log contains unicode and will not be printed")

    _print_info("Running: {0:s}".format(cmd), loglevel="INFO")

    sys.stdout.flush()
    starttime = time.time()
    process = None
    stop_log_printer = Event()

    try:
        process = subprocess.Popen(cmd, shell=True)
        kill_callback = kill_callback or process.kill

        def clock_killer():
            while process.poll() is None and (timeout >= 0):
                currenttime = time.time()
                if currenttime - starttime < timeout:
                    DEBUG and _print_warn(
                        "Clock Reaper: has been running for {0:f}, must finish in {1:f}".format(
                            currenttime - starttime, timeout
                        )
                    )
                else:
                    _print_warn(
                        "Clock Reaper: Timeout reached for '{0:s}'... sending the KILL signal".format(
                            cmd
                        )
                    )
                    kill_callback()
                time.sleep(INTERRUPT_TIME)

        def log_reader(logfile, stop_event):
            bytes_read = prior_log_bytes_read  # skip any previous runs' output
            while not stop_event.isSet():
                if logfile is not None and os.path.exists(logfile):
                    with codecs.open(
                        logfile, "r", encoding="UTF-8", errors="ignore", buffering=0
                    ) as foutlog:
                        foutlog.seek(bytes_read, 0)
                        lines = foutlog.readlines()
                        bytes_read = foutlog.tell()
                        for line in lines:
                            line and _print_info(line)
                time.sleep(LIVELOG_TIME)  # wait for the log to go to disk

        Thread(target=clock_killer, args=tuple([process])).start()
        if log is not None:
            # crucial - child process should not write to stdout unless it is
            # the container process itself
            Thread(target=log_reader, args=tuple([logfile, stop_log_printer])).start()

        while process.poll() is None:
            currenttime = time.time()
            DEBUG and _print_info(
                "God mode on: has been running for {0:f}".format(
                    currenttime - starttime
                )
            )
            # this is probably not ideal as it interrupts the process every few seconds,
            time.sleep(INTERRUPT_TIME)
            # check whether there is an alternative with a callback

        assert hasattr(process, "returncode"), "No returncode after termination!"
    finally:
        stop_log_printer.set()
        if (process is not None) and process.returncode:
            raise StimelaCabRuntimeError(
                "%s: returns errr code %d" % (command, process.returncode)
            )



import os.path
import sys
import inspect
from importlib import metadata
import logging
from logging import StreamHandler
import re
import getpass
import time

from stimela.utils.logger import (
    SelectiveFormatter,
    ColorizingFormatter,
    ConsoleColors,
    MultiplexingHandler,
)

__version__ = metadata.version(__package__)

# Get to know user
try:
    USER = getpass.getuser()
except OSError:
    # The point is to avoid containers with the same name when using there multiple users using docker
    USER = hex(id(time.ctime()))
CAB_USERNAME = re.sub("[^0-9a-zA-Z]+", "_", USER).lower()

root = os.path.dirname(__file__)

CAB_PATH = os.path.join(root, "cargo/cab")
BASE_PATH = os.path.join(root, "cargo/base")


GLOBALS = {"foo": "bar"}
del GLOBALS["foo"]


def register_globals():
    frame = inspect.currentframe().f_back
    frame.f_globals.update(GLOBALS)


# Get base images
# All base images must be on dockerhub
BASE = os.listdir(BASE_PATH)
CAB = list()

for item in os.listdir(CAB_PATH):
    try:
        # These files must exist for a cab image to be valid
        ls_cabdir = os.listdir("{0}/{1}".format(CAB_PATH, item))
        paramfile = "parameters.json" in ls_cabdir
        srcdir = "src" in ls_cabdir
    except OSError:
        continue
    if paramfile and srcdir:
        CAB.append(item)


_logger = None


log_console_handler = log_formatter = log_boring_formatter = log_colourful_formatter = (
    None
)


def is_logger_initialized():
    return _logger is not None


def logger(
    name="STIMELA",
    propagate=False,
    console=True,
    boring=False,
    fmt="{asctime} {name} {levelname}: {message}",
    col_fmt="{asctime} {name} %s{levelname}: {message}%s"
    % (ConsoleColors.BEGIN, ConsoleColors.END),
    sub_fmt="# {message}",
    col_sub_fmt="%s# {message}%s" % (ConsoleColors.BEGIN, ConsoleColors.END),
    datefmt="%Y-%m-%d %H:%M:%S",
    loglevel="INFO",
):
    """Returns the global Stimela logger (initializing if not already done so, with the given values)"""
    global _logger
    if _logger is None:
        _logger = logging.getLogger(name)
        _logger.setLevel(getattr(logging, loglevel))
        _logger.propagate = propagate

        global \
            log_console_handler, \
            log_formatter, \
            log_boring_formatter, \
            log_colourful_formatter

        # this function checks if the log record corresponds to stdout/stderr output from a cab
        def _is_from_subprocess(rec):
            return hasattr(rec, "stimela_subprocess_output")

        log_boring_formatter = SelectiveFormatter(
            logging.Formatter(fmt, datefmt, style="{"),
            [(_is_from_subprocess, logging.Formatter(sub_fmt, datefmt, style="{"))],
        )

        log_colourful_formatter = SelectiveFormatter(
            ColorizingFormatter(col_fmt, datefmt, style="{"),
            [
                (
                    _is_from_subprocess,
                    ColorizingFormatter(
                        fmt=col_sub_fmt,
                        datefmt=datefmt,
                        style="{",
                        default_color=ConsoleColors.DIM,
                    ),
                )
            ],
        )

        log_formatter = log_boring_formatter if boring else log_colourful_formatter

        if console:
            if (
                "SILENT_STDERR" in os.environ
                and os.environ["SILENT_STDERR"].upper() == "ON"
            ):
                log_console_handler = StreamHandler(stream=sys.stdout)
            else:
                log_console_handler = MultiplexingHandler()
            log_console_handler.setFormatter(log_formatter)
            log_console_handler.setLevel(getattr(logging, loglevel))
            _logger.addHandler(log_console_handler)

    return _logger


Recipe = __import__("stimela.recipe", fromlist=["Recipe"]).Recipe



import os
import subprocess
import glob
import yaml
import shlex
import shutil

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

with open(CONFIG, "r") as _std:
    cab = yaml.safe_load(_std)

junk = cab["junk"]

args = []
for param in cab["parameters"]:
    name = param["name"]
    value = param["value"]

    if value in [False, None]:
        continue

    if name in "restoring-beam scale".split() and hasattr(value, "__iter__"):
        value = ",".join(value)

    if value is True:
        value = ""
        if name == "f":
            args.append("-f")
            continue

    # Positional arguments
    if name == "input-image":
        inim = value
        continue

    elif name == "input-skymodel":
        inlsm = value
        continue

    elif name == "output-image":
        outim = value
        continue

    args.append("{0}{1} {2}".format(cab["prefix"], name, value))


_runc = " ".join([cab["binary"]] + args + [inim, inlsm, outim])

try:
    subprocess.check_call(shlex.split(_runc))
finally:
    for item in junk:
        for dest in [
            OUTPUT,
            MSDIR,
        ]:  # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)



import os
import subprocess
import yaml
import glob
import shutil
import shlex

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

with open(CONFIG, "r") as _std:
    cab = yaml.safe_load(_std)

junk = cab["junk"]

mslist = []
field = []

args = []
for param in cab["parameters"]:
    name = param["name"]
    value = param["value"]

    if value in [False, None]:
        continue
    if name == "primary-beam":
        value = "'{}'".format(value)
    if name == "pa-range" and hasattr(value, "__iter__"):
        value = ",".join(value)
    if value is True:
        value = ""
    if name == "pa-from-ms" and hasattr(value, "__iter__"):
        mslist = value
        continue
    if name == "field-id" and hasattr(value, "__iter__"):
        field = value
        continue

    # Positional arguments
    if name == "input-skymodel":
        inlsm = value
        continue
    elif name == "output-skymodel":
        outlsm = value
        continue

    args.append("{0}{1} {2}".format(cab["prefix"], name, value))

if mslist:
    if len(field) == 0:
        field = [0] * len(mslist)
    pa_from_ms = ",".join(["{0}:{1}".format(ms, i) for ms, i in zip(mslist, field)])
    args.append("--pa-from-ms {}".format(pa_from_ms))

_runc = " ".join([cab["binary"]] + args + [inlsm, outlsm])

try:
    subprocess.check_call(shlex.split(_runc))
finally:
    for item in junk:
        for dest in [
            OUTPUT,
            MSDIR,
        ]:  # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)



import pyfits as fitsio
import numpy
import os
import sys
import tempfile
import subprocess
import shlex

imagename = sys.argv[1]
msname = sys.argv[2]
cell = sys.argv[3]

hdulist = fitsio.open(imagename)
if isinstance(hdulist, list):
    hdu = hdulist[0]
else:
    hdu = hdulist

tfile = tempfile.NamedTemporaryFile(suffix=".fits")
tfile.flush()

nchan = hdu.data.shape[0]
sys.stdout.write("Creating template image\n")
os.system("rm -fr {0}".format(tfile.name))
_runc = " ".join(
    ["lwimager"]
    + [
        "ms=" + msname,
        "fits=" + tfile.name,
        "data=psf",
        "npix=64",
        "mode=mfs",
        "nchan={:d}".format(nchan),
        "img_nchan={:d}".format(nchan),
        "prefervelocity=False",
        "cellsize={0}".format(cell),
    ]
)
subprocess.check_call(shlex.split(_runc))

sys.stdout.write(
    "Created template image. Will now proceed to simulate FITS model into MS\n"
)


with fitsio.open(tfile.name) as hdu0:
    header = hdu0[0].header

tfile.close()

naxis = hdu.header["NAXIS"]
raNpix = hdu.header["NAXIS1"]
decNpix = hdu.header["NAXIS2"]

header["CRPIX1"] = raNpix / 2.0
header["CRPIX2"] = decNpix / 2.0

if naxis < 2:
    raise RuntimeError("FITS image has to have at least two axes")

elif naxis == 2:
    hdu.data = hdu.data[numpy.newaxis, numpy.newaxis, ...]
    sys.stdout.write("WARNING::FITS image has 2 axes. Will add STOKES and FREQ axes\n")
elif naxis == 3:
    sys.stdout.write("WARNING::FITS image has 3 axes.\n")
    # Try to find out what is the 3rd axis
    third_axis = hdu.header.get("CTYPE3", None)
    if third_axis is None:
        sys.stdout.write(
            "WARNING:: Third axis is not labelled, assuming its the FREQ Axis\n"
        )
        hdu.data = hdu.data[:, numpy.newaxis, ...]
    elif third_axis.lower() == "freq":
        hdu.data = hdu.data[:, numpy.newaxis, ...]
    elif third_axis.lower() == "stokes":
        hdu.data = hdu.data[numpy.newaxis, ...]
    else:
        sys.stdout.write(
            "WARNING:: CTYPE3 value [{}] is unknown, will ignore it and assume 3rd axis is FREQ\n".format(
                third_axis
            )
        )
        hdu.data = hdu.data[:, numpy.newaxis, ...]

elif naxis == 4:
    sys.stdout.write(
        "FITS image has four axes. If the axes are labeled (via CTYPE<axis index>), will attempt to restructure it so it is fit to predict from\n"
    )
    third_axis = hdu.header.get("CTYPE3", None)
    if third_axis is None:
        raise RuntimeError(
            "ABORT:: FITS axes are not labelled. Cannot proceed. Please label STOKES and FREQ accordingly\n"
        )
    elif third_axis.lower() == "freq":
        hdu.data = numpy.rollaxis(hdu.data, 1)
else:
    sys.stdout.write(
        "WARNING:: FITS image has more than 4 axes. Passing the  buck to lwimger\n"
    )


hdu.header = header

hdulist.writeto(sys.argv[4], clobber=True)



import os
import shlex
import shutil
import subprocess
import glob
from collections import OrderedDict
import json

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

with open(CONFIG, "rb") as stdr:
    cab = json.load(stdr, object_pairs_hook=OrderedDict)

junk = cab["junk"]
args = []

for param in cab["parameters"]:
    name = param["name"]
    value = param["value"]
    dtype = param["dtype"]
    if value is None:
        continue

    if name == "infits":
        infits = value
        continue

    if isinstance(value, list):
        values = map(str, value)
        args += [f"{cab['prefix']}{name} " + ",".join(values)]
    elif name == "output-prefix":
        args += [f"{cab['prefix']}{name} {OUTPUT}/{value}"]
    elif isinstance(dtype, str) and dtype == "bool":
        args += [f"{cab['prefix']}{name}"]
    else:
        args += ["{0}{1} {2}".format(cab["prefix"], name, value)]

binary = f"{os.environ['STIMELA_VENV_BIN']}/{cab['binary']}"
_runc = " ".join([binary] + args + [infits])


subprocess.check_call(shlex.split(f"echo '{_runc}'"))
try:
    subprocess.check_call(shlex.split(_runc))
finally:
    for item in junk:
        for dest in [
            OUTPUT,
            MSDIR,
        ]:  # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)



import os
import subprocess
import shlex
import shutil
import glob
import yaml

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
MSDIR = os.environ["MSDIR"]
OUTPUT = os.environ["OUTPUT"]

with open(CONFIG, "r") as _std:
    cab = yaml.safe_load(_std)
junk = cab["junk"]

args = {}
for param in cab["parameters"]:
    name = param["name"]
    value = param["value"]

    if value is None:
        continue
    args[name] = value

if not os.path.exists(OUTPUT + "/mask_mosaic"):
    os.mkdir(OUTPUT + "/mask_mosaic")

outdir = OUTPUT + "/mask_mosaic"

try:
    make_table = " ".join(["mImgtbl", args["input_dir"], outdir + "/mosaic_table.tbl"])
    subprocess.check_call(shlex.split(make_table))

    make_header = " ".join(
        ["mMakeHdr", outdir + "/mosaic_table.tbl", outdir + "/mosaic_header.hdr"]
    )
    subprocess.check_call(shlex.split(make_header))

    project_mosaic = " ".join(
        [
            "mProjExec",
            "-p",
            args["input_dir"],
            outdir + "/mosaic_table.tbl",
            outdir + "/mosaic_header.hdr",
            outdir,
            outdir + "/stats.tbl",
        ]
    )
    subprocess.check_call(shlex.split(project_mosaic))

    make_mosaic_table = ["mImgtbl", outdir, outdir + "/mosaic_table2.tbl"]
    subprocess.check_call(shlex.split(make_mosaic_table))

    _runc = " ".join(
        [
            "mAdd",
            "-p",
            args["input_dir"],
            outdir + "/mosaic_table2.tbl",
            outdir + "/mosaic_header.hdr",
            OUTPUT + "/mosaic.fits",
        ]
    )
    subprocess.check_call(shlex.split(_runc))
finally:
    for item in junk:
        for dest in [
            OUTPUT,
            MSDIR,
        ]:  # these are the only writable volumes in the container
            items = glob.glob("{dest}/{item}".format(**locals()))
            for f in items:
                if os.path.isfile(f):
                    os.remove(f)
                elif os.path.isdir(f):
                    shutil.rmtree(f)
                # Leave other types



import stimela
from stimela.pathformatter import pathformatter as spf
import unittest


class TestMkReduce(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        unittest.TestCase.setUpClass()
        # I/O
        global INPUT
        INPUT = "input"
        global MSDIR
        MSDIR = "msdir"

        global MS
        MS = "1491291289.1GC.ms"
        global PREFIX
        PREFIX = "deep2"

        global LABEL
        LABEL = "test_mkreduction"
        global OUTPUT
        OUTPUT = "output_%s" % LABEL

        stimela.register_globals()

    @classmethod
    def tearDownClass(cls):
        unittest.TestCase.tearDownClass()
        # global OUTPUT
        # shutil.rmtree(OUTPUT)

    def tearDown(self):
        unittest.TestCase.tearDown(self)

    def setUp(self):
        unittest.TestCase.setUp(self)

    def test_end_to_end_reduction(self):
        global INPUT, OUTPUT, MSDIR, MS, LABEL
        recipe = stimela.Recipe(
            "Test reduction script", ms_dir=MSDIR, JOB_TYPE="docker", log_dir="logs"
        )

        imname1 = "deep2.1gc"
        recipe.add(
            "cab/ddfacet",
            "ddfacet_test1",
            {
                "Data-MS": [MS],
                "Output-Name": imname1,
                "Image-NPix": 2048,
                "Image-Cell": 2,
                "Cache-Reset": True,
                "Freq-NBand": 2,
                "Freq-NDegridBand": 4,
                "Weight-ColName": "WEIGHT",
                "Data-ChunkHours": 0.1,
                "Data-Sort": True,
                "Log-Boring": True,
                "Deconv-MaxMajorIter": 2,
                "Deconv-MaxMinorIter": 1500,
                "Predict-ColName": "MODEL_DATA",
            },
            input=INPUT,
            output=OUTPUT,
            shared_memory="8gb",
            label="image1",
            time_out=1800,
        )

        recipe.add(
            "cab/tricolour",
            "flag_data",
            {
                "ms": MS,
                "data-column": "DATA",
                "window-backend": "numpy",
                "flagging-strategy": "total_power",
                "subtract-model-column": "MODEL_DATA",
            },
            input=INPUT,
            output=OUTPUT,
            label="flag_data",
            time_out=1800,
        )

        maskname0 = "MASK.fits"
        recipe.add(
            "cab/cleanmask",
            "mask0",
            {
                "image": "%s.app.restored.fits:output" % (imname1),
                "output": "%s:output" % (maskname0),
                "dilate": False,
                "sigma": 25,
            },
            input=INPUT,
            output=OUTPUT,
            label="mask0:: Make mask",
            time_out=1800,
        )

        recipe.add(
            "cab/ddfacet",
            "ddfacet_test2",
            {
                "Data-MS": [MS],
                "Output-Name": imname1,
                "Output-Images": "DdPAMRrIikze",
                "Image-NPix": 4096,
                "Image-Cell": 2.0,
                "Cache-Reset": True,
                "Freq-NBand": 2,
                "Freq-NDegridBand": 4,
                "Mask-External": "%s:output" % (maskname0),
                "Weight-ColName": "WEIGHT",
                "Data-ChunkHours": 0.1,
                "Data-Sort": True,
                "Log-Boring": True,
                "Deconv-MaxMajorIter": 2,
                "Deconv-MaxMinorIter": 1500,
            },
            input=INPUT,
            output=OUTPUT,
            shared_memory="24gb",
            label="image2",
            time_out=1800,
        )

        recipe.add(
            "cab/shadems",
            "shadems_test",
            {
                "ms": MS,
                "xaxis": "DATA:imag",
                "yaxis": "real",
                "col": "DATA",
                "png": "%s_shadems_test_real_imag" % (PREFIX),
            },
            input=INPUT,
            output=OUTPUT,
            label="shadems_test",
            time_out=1800,
        )

        # # First selfcal round
        recipe.add(
            "cab/catdagger",
            "auto_tagger_{}_{}".format("decaltest", "1"),
            {
                "ds9-reg-file": "{}.{}.dE.reg".format("decaltest", "1"),
                "ds9-tag-reg-file": "{}.{}.dE.clusterleads.reg".format(
                    "decaltest", "1"
                ),
                "sigma": 10,
                "min-distance-from-tracking-centre": 350,
                "noise-map": "{}.app.residual.fits:output".format(imname1),
            },
            input=INPUT,
            output=OUTPUT,
            label="auto_tagger_{}_{}".format("decaltest", "1"),
            shared_memory="250g",
        )

        recipe.add(
            "cab/cubical_ddf",
            "cubical_cal",
            {
                "data-ms": MS,
                "data-column": "DATA",
                "dist-nworker": 4,
                "dist-nthread": 1,
                "dist-max-chunks": 20,
                "data-freq-chunk": 0,
                "data-time-chunk": 1,
                "model-list": spf(
                    "MODEL_DATA+-{{}}{}@{{}}{}:{{}}{}@{{}}{}".format(
                        imname1 + ".DicoModel",
                        "{}.{}.dE.reg".format("decaltest", "1"),
                        imname1 + ".DicoModel",
                        "{}.{}.dE.reg".format("decaltest", "1"),
                    ),
                    "output",
                    "output",
                    "output",
                    "output",
                ),
                "log-verbose": "solver=0",
                "weight-column": "WEIGHT",
                "flags-apply": "FLAG",
                "flags-auto-init": "legacy",
                "madmax-enable": False,
                "madmax-threshold": [0, 0, 10],
                "madmax-global-threshold": [0, 0],
                "sol-jones": "g,dd",
                "sol-stall-quorum": 0.95,
                "out-name": "cubicaltest",
                "out-column": "CORRECTED_DATA",
                "g-type": "complex-2x2",
                "g-freq-int": 0,
                "dd-freq-int": 0,
                "g-time-int": 20,
                "dd-time-int": 20,
                "g-max-iter": 10,
                "sol-term-iters": 10,
                "g-update-type": "phase-diag",
                "dd-update-type": "complex-2x2",
                "out-subtract-dirs": "1:",
                "dd-fix-dirs": "0",
                "dd-max-iter": 200,
                "dd-clip-high": 0,
                "dd-clip-low": 0,
                "dd-max-prior-error": 0.35,
                "dd-max-post-error": 0.35,
                "degridding-NDegridBand": 3,
                "degridding-MaxFacetSize": 0.15,
                "out-mode": "sr",
            },
            input=INPUT,
            output=OUTPUT,
            label="cubical",
            shared_memory="24gb",
            time_out=3600,
        )

        recipe.add(
            "cab/cubical",
            "cubical_cal2",
            {
                "data-ms": MS,
                "data-column": "DATA",
                "dist-nworker": 4,
                "dist-nthread": 1,
                "dist-max-chunks": 20,
                "data-freq-chunk": 0,
                "data-time-chunk": 1,
                "model-list": spf("MODEL_DATA"),
                "weight-column": "WEIGHT",
                "flags-apply": "FLAG",
                "flags-auto-init": "legacy",
                "madmax-enable": False,
                "madmax-threshold": [0, 0, 10],
                "madmax-global-threshold": [0, 0],
                "sol-jones": "g",
                "sol-stall-quorum": 0.95,
                "out-name": "cubicaltest",
                "out-column": "CORRECTED_DATA",
                "log-verbose": "solver=2",
                "g-type": "complex-2x2",
                "g-freq-int": 0,
                "g-time-int": 20,
                "g-max-iter": 10,
                "sol-term-iters": 10,
                "out-overwrite": True,
                "g-update-type": "complex-2x2",
            },
            input=INPUT,
            output=OUTPUT,
            label="cubical",
            shared_memory="24gb",
            time_out=1800,
        )

        recipe.add(
            "cab/ragavi_vis",
            "ragavi_vis_test",
            {
                "ms": MS,
                "xaxis": "imaginary",
                "yaxis": "real",
                "data-column": "CORRECTED_DATA",
                "htmlname": "%s_ragavi_vis_real_imag" % (PREFIX),
            },
            input=INPUT,
            output=OUTPUT,
            label="ragavi_vis_test",
            time_out=1800,
        )
        recipe.run()



import os
import Crasa.Crasa as crasa
import yaml
import glob
import shutil

CONFIG = os.environ["CONFIG"]
INPUT = os.environ["INPUT"]
OUTPUT = os.environ["OUTPUT"]
MSDIR = os.environ["MSDIR"]

with open(CONFIG, "r") as _std:
    cab = yaml.safe_load(_std)
junk = cab["junk"]

unstack_params = ["unstack", "nchans", "keep_casa_images", "port2fits"]
unstack_args = {}
immath_args = {}


def rm_fr(item):
    os.system("rm -fr {}".format(item))


for param in cab["parameters"]:
    name = param["name"]
    value = param["value"]

    if value is None:
        continue

    if name in unstack_params:
        unstack_args[name] = value
    else:
        immath_args[name] = value

unstack = unstack_args.pop(cab["binary"], False)

if not unstack:
    task = crasa.CasaTask("immath", **immath_args)
    try:
        task.run()
    finally:
        for item in junk:
            for dest in [
                OUTPUT,
                MSDIR,
            ]:  # these are the only writable volumes in the container
                items = glob.glob("{dest}/{item}".format(**locals()))
                for f in items:
                    if os.path.isfile(f):
                        os.remove(f)
                    elif os.path.isdir(f):
                        shutil.rmtree(f)
else:
    images = immath_args["imagename"]
    for image in images:
        for i in range(unstack_args["nchans"]):
            ext = image.split(".")[-1]
            chan_num = str(i)
            outfile = "{:s}-{:s}.{:s}".format(immath_args["outfile"], chan_num, ext)
            run_immath_args = immath_args.copy()
            run_immath_args["imagename"] = image
            run_immath_args["outfile"] = outfile
            run_immath_args["chans"] = chan_num
            task = crasa.CasaTask(cab["binary"], **run_immath_args)
            task.run()
            if unstack_args["port2fits"]:
                print("Converting CASA images to FITS images")
                fits = outfile + ".fits"
                task = crasa.CasaTask(
                    "exportfits",
                    **dict(imagename=outfile, fitsimage=fits, overwrite=True),
                )
                task.run()
                if not unstack_args["keep_casa_images"]:
                    rm_fr(outfile)



import os
import sys
from pyrap.tables import table
import Cattery
from scabha import config, parameters_dict, prun, OUTPUT, log

CODE = os.path.join(os.environ["STIMELA_MOUNT"], "code")
CONFIG = os.environ["CONFIG"]

binary = config.binary

jdict = {}
for name, value in parameters_dict.items():
    if value is None:
        continue
    if value is False:
        value = 0
    elif value is True:
        value = 1

    jdict[name] = value

msname = jdict["msname"]
THREADS = jdict.pop("threads", 2)
msbase = os.path.basename(msname)[:-3]
prefix = jdict.pop("prefix", None) or "{0}/{1}".format(OUTPUT, msbase)
params = {}

# options for writing flags
writeflags = jdict.pop("write-flags-to-ms", None)
if writeflags:
    params["ms_sel.ms_write_flags"] = 1
    params["ms_sel.ms_fill_legacy_flags"] = (
        1 if jdict.pop("fill-legacy-flags", False) else 0
    )

write_flagset = jdict.pop("write-flagset", None)
if write_flagset:
    params["ms_wfl.write_bitflag"] = write_flagset
    params["ms_sel.ms_write_flag_policy"] = (
        "'add to set'" if jdict.pop("write-flag-policy", False) else "'replace set'"
    )

# Read flags options
readflagsets = jdict.pop("read-flagsets", False)
if readflagsets:
    params["ms_rfl.read_flagsets"] = readflagsets

params["ms_sel.ms_read_flags"] = 1 if jdict.pop("read-flags-from-ms", False) else 0
params["ms_rfl.read_legacy_flags"] = 1 if jdict.pop("read-legacy-flags", False) else 0

params["ms_sel.msname"] = msname
field_id = jdict.pop("field-id", 0)
spw_id = jdict.pop("spw-id", 0)
params["ms_sel.tile_size"] = jdict.pop("tile-size", 16)
params["ms_sel.ddid_index"] = spw_id
params["ms_sel.field_index"] = field_id

TDL = jdict.pop("tdlconf", None) or CODE + "/tdlconf.profiles"
SECTION = jdict.pop("section", None) or "stefcal"
skymodel = jdict.pop("skymodel", None)
beam_files_pattern = jdict.pop("beam-files-pattern", False)
jones_type = jdict.pop("jones-implementation", "Gain2x2")

column = jdict.pop("column", "DATA")
outcol = jdict.pop("output-column", "CORRECTED_DATA")

params["ms_sel.input_column"] = column
params["ms_sel.output_column"] = outcol
params["tiggerlsm.filename"] = skymodel
params["tiggerlsm.lsm_subset"] = jdict.get("subset", "all")
params["do_output"] = jdict.pop("output-data", "CORR_RES")
saveconf = jdict.pop("save-config", None)
params["ms_sel.ms_corr_sel"] = "'{}'".format(jdict.pop("correlations", "2x2"))

label = jdict.pop("label", None)

model_column = jdict.pop("model-column", "MODEL_DATA")

gjones = jdict.pop("Gjones", False)
if gjones:
    time_smooth, freq_smooth = jdict.get("Gjones-smoothing-intervals", (1, 1))
    time_int, freq_int = jdict.get("Gjones-solution-intervals", (1, 1))
    mode = "apply" if jdict.get("Gjones-apply-only", False) else "solve-save"

    gjones_gains = jdict.pop("Gjones-gain-table", None) or "{0}/{1}{2}.gain.cp".format(
        OUTPUT, msbase, "-%s" % label if label else ""
    )
    params.update(
        {
            "stefcal_gain.enabled": 1,
            "stefcal_gain.mode": mode,
            "stefcal_gain.reset": 0 if mode == "apply" else 1,
            "stefcal_gain.timeint": time_int,
            "stefcal_gain.freqint": freq_int,
            "stefcal_gain.flag_ampl": jdict.get("Gjones-ampl-clipping", 0),
            "stefcal_gain.flag_chisq": jdict.get("Gjones-chisq-clipping", 0),
            "stefcal_gain.flag_chisq_threshold": jdict.get("Gjones-thresh-sigma", 10),
            "stefcal_gain.flag_ampl_low": jdict.get("Gjones-ampl-clipping-low", 0.3),
            "stefcal_gain.flag_ampl_high": jdict.get("Gjones-ampl-clipping-high", 2),
            "stefcal_gain.implementation": jdict.get("Gjones-matrix-type", "Gain2x2"),
            "stefcal_gain.table": gjones_gains,
        }
    )

bjones = jdict.pop("Bjones", False)
if bjones:
    time_smooth, freq_smooth = jdict.get("Bjones-smoothing-intervals", (1, 0))
    time_int, freq_int = jdict.get("Bjones-solution-intervals", (1, 0))
    mode = "apply" if jdict.get("Bjones-apply-only", False) else "solve-save"

    bjones_gains = jdict.pop("Bjones-gain-table", None) or "{0}/{1}{2}.gain1.cp".format(
        OUTPUT, msbase, "-%s" % label if label else ""
    )
    params.update(
        {
            "stefcal_gain1.enabled": 1,
            "stefcal_gain1.label": "B",
            "stefcal_gain1.mode": mode,
            "stefcal_gain1.reset": 0 if mode == "apply" else 1,
            "stefcal_gain1.timeint": time_int,
            "stefcal_gain1.freqint": freq_int,
            "stefcal_gain1.flag_ampl": jdict.get("Bjones-ampl-clipping", 0),
            "stefcal_gain1.flag_chisq": jdict.get("Bjones-chisq-clipping", 0),
            "stefcal_gain1.flag_chisq_threshold": jdict.get("Bjones-thresh-sigma", 10),
            "stefcal_gain1.flag_ampl_low": jdict.get("Bjones-ampl-clipping-low", 0.3),
            "stefcal_gain1.flag_ampl_high": jdict.get("Bjones-ampl-clipping-high", 2),
            "stefcal_gain1.implementation": jdict.get("Bjones-matrix-type", "Gain2x2"),
            "stefcal_gain1.table": bjones_gains,
        }
    )

beam = jdict.pop("Ejones", False)
if beam and beam_files_pattern:
    params.update(
        {
            "me.e_enable": 1,
            "me.p_enable": 1,
            "me.e_module": "Siamese_OMS_pybeams_fits",
            "me.e_all_stations": 1,
            "pybeams_fits.sky_rotation": 1
            if jdict.pop("parallactic-angle-rotation", False)
            else 0,
            "pybeams_fits.l_axis": jdict.pop("beam-l-axis", "L"),
            "pybeams_fits.m_axis": jdict.pop("beam-m-axis", "M"),
            "pybeams_fits.filename_pattern": "'{}'".format(beam_files_pattern),
        }
    )


ddjones = jdict.pop("DDjones", False)
if ddjones:
    time_int, freq_int = jdict.pop("DDjones-solution-intervals", (1, 1))
    time_smooth, freq_smooth = jdict.pop("DDjones-smoothing-intervals", (1, 1))

    mode = "apply" if jdict.get("DDjones-apply-only", False) else "solve-save"

    ddjones_gains = jdict.pop(
        "DDjones-gain-table", None
    ) or "{0}/{1}{2}.diffgain.cp".format(OUTPUT, msbase, "-%s" % label if label else "")
    params.update(
        {
            "stefcal_diffgain.enabled": 1,
            "stefcal_diffgain.reset": 0 if mode == "apply" else 1,
            "stefcal_diffgain.flag_ampl": 0,
            "stefcal_diffgain.flag_chisq": 1,
            "stefcal_diffgain.flag_chisq_threshold": 5,
            "stefcal_diffgain.freqint": freq_int,
            "stefcal_diffgain.freqsmooth": freq_smooth,
            "stefcal_diffgain.label": jdict.pop("DDjones-tag", "dE"),
            "stefcal_diffgain.max_diverge": 1,
            "stefcal_diffgain.mode": mode,
            "stefcal_diffgain.niter": 50,
            "stefcal_diffgain.omega": 0.5,
            "stefcal_diffgain.quota": 0.95,
            "stefcal_diffgain.table": ddjones_gains,
            "stefcal_diffgain.timeint": time_int,
            "stefcal_diffgain.timesmooth": time_smooth,
            "stefcal_diffgain.implementation": jdict.get(
                "DDjones-matrix-type", "Gain2x2"
            ),
        }
    )

ifrjones = jdict.pop("DDjones", False)
if ifrjones:
    ifrjones_gains = jdict.pop(
        "IFRjones-gain-table", None
    ) or "{0}/{1}{2}.ifrgain.cp".format(OUTPUT, msbase, "-%s" % label if label else "")
    mode = "apply" if jdict.get("IFRjones-apply-only", False) else "solve-save"

    params.update(
        {
            "stefcal_ifr_gain_mode": mode,
            "stefcal_ifr_gains": 1,
            "stefcal_ifr_gain_reset": 0 if mode == "apply" else 1,
            "stefcal_reset_ifr_gains": 0 if mode == "apply" else 1,
            "stefcal_ifr_gain.table": ifrjones_gains,
        }
    )

makeplots = jdict.pop("make-plots", False)

gjones_plotprefix = prefix + "-gjones_plots"
bjones_plotprefix = prefix + "-bjones_plots"
ddjones_plotprefix = prefix + "-ddjones_plots"
ifrjones_plotprefix = prefix + "-ifrjones_plots"


def run_meqtrees(msname):
    prefix = ["--mt %d -c %s [%s]" % (THREADS, TDL, SECTION)]
    CATTERY_PATH = os.path.dirname(Cattery.__file__)
    suffix = ["%s/Calico/calico-stefcal.py =stefcal" % CATTERY_PATH]
    options = {}
    options.update(params)
    if jdict.pop("add-vis-model", 0):
        options["read_ms_model"] = 1
        options["ms_sel.model_column"] = model_column

    taql = jdict.get("data-selection", None)
    if taql:
        options["ms_sel.ms_taql_str"] = taql

    args = []
    for key, value in options.items():
        if isinstance(value, str) and value.find(" ") > 0:
            value = '"{:s}"'.format(value)
        args.append("{0}={1}".format(key, value))

    args = prefix + args + suffix

    _runc = " ".join([binary] + args + ["-s {}".format(saveconf) if saveconf else ""])
    if prun(_runc) != 0:
        sys.exit(1)

    log.info("MeqTrees Done!")
    # now plot the gains
    if makeplots:
        log.info("Preparing to make gain plots")
        plotgains = __import__("Owlcat.Gainplots")

        feed_tab = table(msname + "/FEED")
        log.info("Extracting feed type from MS")
        feed_type = set(feed_tab.getcol("POLARIZATION_TYPE")["array"])
        feed_type = "".join(feed_type)
        log.info("Feed type is [%s]" % feed_type)

        if feed_type.upper() in ["XY", "YX"]:
            feed_type = "XY"
        else:
            feed_type = "RL"

        if gjones:
            log.info("Making Gain plots (G)...")
            plotgains.make_gain_plots(
                gjones_gains, prefix=gjones_plotprefix, feed_type=feed_type
            )

        if bjones:
            log.info("Making Gain plots (B)...")
            plotgains.make_gain_plots(
                bjones_gains,
                gain_label="B",
                prefix=bjones_plotprefix,
                feed_type=feed_type,
            )

        if ddjones:
            log.info("Making differential gain plots...")
            plotgains.make_diffgain_plots(
                ddjones_gains, prefix=ddjones_plotprefix, feed_type=feed_type
            )

        if ifrjones:
            log.info("Making IFR gain plots...")
            plotgains.make_ifrgain_plots(
                ifrjones_gains, prefix=ifrjones_plotprefix, feed_type=feed_type
            )


run_meqtrees(msname)
